/**
 * Simple localStorage wrapper
 */
Ember.LocalStorage = Ember.Object.extend({
  id: 'ember-default-local-storage',

  data: function(key, value) {
    var id = this.get('id');

    if (value === undefined) {
      return JSON.parse(localStorage.getItem(id));
    } else {
      localStorage.setItem(id, JSON.stringify(value));
      return value;
    }

  // Note how we use `cacheable` to minimise deserialization
  }.property().cacheable()

});

/**
 * Application
 */
var App = Ember.Application.create({

  // this huck will be executed on DOMLoad
  ready: function() {
    App.projectsController.loadData();
  },

  // create LocalStorage instance
  localStorage: Ember.LocalStorage.create({
    id: 'projects-manager-data'
  })
});

/**
 * Models
 */

App.Project = Ember.Object.extend({

  decrementPriority: function() {
    // incrementProperty and decrementPriority
    // are helpers methods available on any Ember.Object
    this.decrementProperty('priority');
  },
  incrementPriority: function() {
    this.incrementProperty('priority');
  },

  // we override destroy for this object, to ensure
  // it is removed from collection
  destroy: function() {
    App.projectsController.removeObject(this);

    this._super();
  },

  toJSON: function() {
    // getProperties is another helper on Ember.Object
    // this will return a literal object ready to be serialized
    return this.getProperties('name', 'description', 'priority');
  },

  isNew: function() {
    // check if this project is part of collection
    return !App.projectsController.contains(this);
  }.property()
});

// here we add class level method
App.Project.reopenClass({
  createWithPriority: function(priority) {
    return App.Project.create({
      priority: priority
    });
  }
});

/**
 * Controllers
 */

App.projectsController = Ember.ArrayController.create({

  // this is a new project
  item: App.Project.createWithPriority(5),
  // this is our actual collection
  // for wich ArrayController serves as a proxy
  // any Enumerable interface method called on this controller
  // will manipulate content array
  content: [],

  // add and save new project
  save: function(evt) {
    evt.preventDefault();

    // push it in the collection
    this.addObject(this.get('item'));

    // and call `saveData` to serialize it and save in localStorage
    this.saveData();

    this.reset();
  },

  // set a project as courent editable item
  // and save a snapshot to allow rollback
  edit: function(project) {
    this.setProperties({
      item: project,
      snapshot: project.toJSON()
    });
  },

  // apply snapshot and reset current item
  cancel: function() {
    var project = this.get('item'),
        snapshot = this.get('snapshot');

    if (project && snapshot) {
      project.setProperties(snapshot);
    }

    this.reset();
  },

  // set current item to a new instance of project
  reset: function() {
    this.setProperties({
      item: App.Project.createWithPriority(5),
      snapshot: null
    });
  },

  // invoke is a helper on Enumerable interface
  // that will call given method on every item
  // this is a shorthand for
  // this.map(function(item) { return item.toJSON(); });
  toJSON: function() {
    return this.invoke('toJSON');
  },

  fromJSON: function(data) {
    var projects = (data || []).map(function(project) {
      return App.Project.create(project);
    });

    // set content array
    this.set('content', projects);
  },

  // serialize and save
  saveData: function() {
    App.localStorage.set('data', this.toJSON());
  },

  // deserialize and load
  loadData: function() {
    this.fromJSON(App.localStorage.get('data'));

    // add an observer on "priority property of each item in given collection"
    // this will save our data on every change to priority property
    // this will be called also if a project is removed from collection
    // so we do not need to call saveData after we call `removeObject` earlier
    this.addObserver('@each.priority', this, 'saveData');
  }
});

/**
 * Views
 */

App.ProjectView = Ember.View.extend({
  // respond to user click on the table row
  click: function(evt) {
    if ($(evt.target).is('td')) {
      App.projectsController.edit(this.get('content'));
    }
  }
});

// add require attribute support to Ember.TextField
Ember.TextField.reopen({
  attributeBindings: ['required'],
  required: false
});

// Add `btn` class and default template to Ember.Button
App.Button = Ember.Button.extend({
  classNames: ['btn'],
  title: null,
  defaultTemplate: Ember.Handlebars.compile('{{title}}')
});

// create some Handlebars hepers
Ember.Handlebars.registerHelper('button', function(options) {
  return Ember.Handlebars.ViewHelper.helper(this, "App.Button", options);
});

Ember.Handlebars.registerHelper('textfield', function(options) {
  return Ember.Handlebars.ViewHelper.helper(this, "Ember.TextField", options);
});

Ember.Handlebars.registerHelper('textarea', function(options) {
  return Ember.Handlebars.ViewHelper.helper(this, "Ember.TextArea", options);
});

